@@ -16,7 +16,6 @@ import glob
 import io
 import logging
 import os
-import pprint
 import re
 import tempfile
 import time
@@ -24,9 +23,10 @@ from collections import namedtuple
 from datetime import datetime
 from functools import lru_cache
 from operator import attrgetter
+
+from ibm_watson_machine_learning.wml_client_error import WMLClientError
 from packaging import version
 
-import requests
 from ibm_watson_machine_learning import APIClient
 
 
@@ -86,6 +86,7 @@ class _CredentialsProvider:
     'credentials' attribute."""
 
     ENVIRONMENT_VARIABLE_NAME = 'DOWML_CREDENTIALS'
+    ENVIRONMENT_VARIABLE_NAME_FILE = 'DOWML_CREDENTIALS_FILE'
     # The keys in the credentials
     APIKEY = 'apikey'
     TOKEN = 'token'
@@ -95,25 +96,26 @@ class _CredentialsProvider:
     COS_CRN = 'cos_resource_crn'
     ML_CRN = 'ml_instance_crn'
 
-    def __init__(self, wml_credentials_file=None, wml_credentials_str = None):
+    def __init__(self, wml_credentials_file=None, wml_credentials_str=None):
         self._logger = logging.getLogger(self.__class__.__name__)
 
         if wml_credentials_str is None:
             if wml_credentials_file is not None:
                 wml_credentials_str = self._read_wml_credentials_from_file(wml_credentials_file)
             else:
-                wml_credentials_str = self._read_wml_credentials_from_env(self.ENVIRONMENT_VARIABLE_NAME)
-            self._logger.debug(f'Found credential string.')
+                wml_credentials_str = self._read_wml_credentials_from_env()
+            self._logger.debug('Found credential string.')
         self.credentials = self.check_credentials(wml_credentials_str)
 
     def usage(self):
-        print(f'It should contain credentials as a Python dict of the form:')
-        print(f'{{\'{self.APIKEY}\': \'<apikey>\', \'{self.URL}\': \'https://us-south.ml.cloud.ibm.com\'}}')
+        print(f'${self.ENVIRONMENT_VARIABLE_NAME} should contain credentials as a Python dict of the form:')
+        print(f'  {{\'{self.APIKEY}\': \'<apikey>\', \'{self.URL}\': \'https://us-south.ml.cloud.ibm.com\'}}')
+        print(f'Or set ${self.ENVIRONMENT_VARIABLE_NAME_FILE} to the path to a file containing the same information.')
 
     def check_credentials(self, wml_cred_str):
         assert type(wml_cred_str) is str
         if len(wml_cred_str) == 0:
-            self._logger.error(f'WML credentials must not be an empty string.')
+            self._logger.error('WML credentials must not be an empty string.')
             self.usage()
             raise InvalidCredentials
         wml_credentials = eval(wml_cred_str)
@@ -127,23 +129,28 @@ class _CredentialsProvider:
         assert type(wml_credentials[self.URL]) is str
         url = wml_credentials[self.URL]
         if len(url) == 0:
-            self._logger.error(f'URL must not be an empty string.')
+            self._logger.error('URL must not be an empty string.')
             raise InvalidCredentials
         if url[-1] == '/':
-            self._logger.warning(f'URL should not have a \'/\' at the end.')
+            self._logger.warning('URL should not have a \'/\' at the end.')
             wml_credentials[self.URL] = url[:-1]
-        self._logger.debug(f'Credentials have the expected structure.')
+        self._logger.debug('Credentials have the expected structure.')
         return wml_credentials
 
-    def _read_wml_credentials_from_env(self, var_name):
+    def _read_wml_credentials_from_env(self):
         """Return a string of credentials suitable for WML from the environment
 
         Raises InvalidCredentials if anything is wrong."""
+        var_name = self.ENVIRONMENT_VARIABLE_NAME
+        var_file_name = self.ENVIRONMENT_VARIABLE_NAME_FILE
         self._logger.debug(f'Looking for credentials in environment variable {var_name}...')
-        try:
+        if var_name in os.environ:
             wml_cred_str = os.environ[var_name]
-        except KeyError:
-            print(f'Environment variable ${var_name} not found.')
+        elif var_file_name in os.environ:
+            self._logger.debug(f'Looking for credentials file name in environment variable {var_file_name}...')
+            wml_cred_str = self._read_wml_credentials_from_file(os.environ[var_file_name])
+        else:
+            print(f'Environment variables ${var_name} or ${var_file_name} not found.')
             self.usage()
             raise InvalidCredentials
 
@@ -186,6 +193,20 @@ class DOWMLLib:
             space_id: the id of the space that should be used. If specified, this
             replaces the one in the credentials
             tz: timezone to use to display time, defaults to Python's default timezone"""
+
+        self.model_type = self.MODEL_TYPES[0]
+        """Type of model to solve"""
+        self.tshirt_size = self.TSHIRT_SIZES[0]
+        """Tee-shirt size for the hardware to solve the model"""
+        self.do_version = self.DO_VERSION
+        """DO engines version to use"""
+        self.timelimit = None
+        """Time limit for the solve, in seconds"""
+        self.inputs = 'assets'
+        """Type of inputs that the created jobs should use"""
+        self.outputs = 'inline'
+        """Type of outputs that the created jobs should use"""
+
         self._logger = logging.getLogger(self.__class__.__name__)
 
         cred_provider = _CredentialsProvider(wml_credentials_file)
@@ -193,11 +214,11 @@ class DOWMLLib:
 
         # A space name in the credentials changes the default
         if cred_provider.SPACE_NAME in wml_credentials:
-            self._logger.debug(f'They contain a space name.')
+            self._logger.debug('They contain a space name.')
             self.space_name = wml_credentials[cred_provider.SPACE_NAME]
 
         if cred_provider.SPACE_ID in wml_credentials:
-            self._logger.debug(f'And they contain a space id.')
+            self._logger.debug('And they contain a space id.')
 
         # The space_id specified here takes precedence
         # over the one, if any, defined in the credentials
@@ -210,28 +231,50 @@ class DOWMLLib:
         # expensive operation.
         self._client = None
         self._space_id = None
-        self.model_type = self.MODEL_TYPES[0]
-        self.tshirt_size = self.TSHIRT_SIZES[0]
-        self.do_version = self.DO_VERSION
-        self.timelimit = None
-        self.inline = False
         self._data_connection = None
         self.tz = tz
 
+    @property
+    def inline(self):
+        self._logger.warning('Attribute \'inline\' is deprecated: use \'inputs\' instead.')
+        return self.inputs == 'inline'
+
+    @inline.setter
+    def inline(self, value):
+        self._logger.warning('Attribute \'inline\' is deprecated: use \'inputs\' instead.')
+        self.inputs = 'inline' if value else 'assets'
+
     def _create_client(self):
         """Create the Python APIClient instance"""
         assert self._client is None
-        self._logger.debug(f'Creating the WML client...')
+        self._logger.debug('Creating the WML client...')
         # http://ibm-wml-api-pyclient.mybluemix.net/#api-for-ibm-cloud
         client = APIClient(self._wml_credentials)
         self._logger.info(f'Creating the client succeeded.  Client version is {client.version}')
         return client
 
+    def _set_default_space(self):
+        space_id_key = _CredentialsProvider.SPACE_ID
+        if space_id_key in self._wml_credentials:
+            space_id = self._wml_credentials[space_id_key]
+            self._logger.debug(f'Using specified space \'{space_id}\'.')
+        else:
+            space_id = self._find_or_create_space()
+
+        self._logger.debug('Setting default space...')
+        self._client.set.default_space(space_id)
+        self._space_id = space_id
+        self._logger.debug('Done.')
+
     def _get_or_make_client(self):
         if self._client is None:
             self._client = self._create_client()
-        if self._space_id is None:
-            self._space_id = self._get_space_id()
+            # The client is pretty much useless when it doesn't yet have a
+            # default space. So let's set it immediately.
+            self._set_default_space()
+        # It would seem natural to assert that self._space_id is not None.
+        # But this fails when we are in unit-tests and we just set _client to
+        # a mock object from outside, without also setting the _space_id.
         return self._client
 
     def solve(self, paths):
@@ -262,7 +305,7 @@ class DOWMLLib:
         try:
             outputs = job_details['entity']['decision_optimization']['output_data']
         except KeyError:
-            self._logger.warning(f'No output structure available for this job')
+            self._logger.warning('No output structure available for this job')
             return None
         for output_data in outputs:
             if output_data['id'] == 'log.txt':
@@ -300,7 +343,7 @@ class DOWMLLib:
         try:
             outputs = details['entity']['decision_optimization']['output_data']
         except KeyError:
-            self._logger.warning(f'No output structure available for this job')
+            self._logger.warning('No output structure available for this job')
             return result
         for output_data in outputs:
             name = output_data['id']
@@ -321,7 +364,8 @@ class DOWMLLib:
                 result[name] = content
         return result
 
-    def _extract_csv_file(self, output_data, tabular_as_csv):
+    @staticmethod
+    def _extract_csv_file(output_data, tabular_as_csv):
         if tabular_as_csv:
             content = io.StringIO()
             writer = csv.writer(content)
@@ -335,7 +379,8 @@ class DOWMLLib:
                                        columns=output_data['fields'])
         return content
 
-    def _extract_regular_file(self, output_data):
+    @staticmethod
+    def _extract_regular_file(output_data):
         content = output_data['content']
         content = content.encode('UTF-8')
         content = base64.b64decode(content)
@@ -350,19 +395,20 @@ class DOWMLLib:
         :return: The job details
         """
         client = self._get_or_make_client()
-        self._logger.debug(f'Fetching output...')
+        self._logger.debug('Fetching output...')
         output_filter = None
         if not with_contents:
             output_filter = 'solve_parameters,solve_state,status'
         elif with_contents == 'log':
             output_filter = 'output_data'
         job_details = self.client_get_job_details(client, job_id, output_filter)
-        self._logger.debug(f'Done.')
+        self._logger.debug('Done.')
         if with_contents != 'full' and with_contents != 'log':
             self.filter_large_chunks_from_details(job_details)
         return job_details
 
-    def client_get_job_details(self, client, job_id, with_filter=None):
+    @staticmethod
+    def client_get_job_details(client, job_id, with_filter=None):
         global the_filter
         global the_old_params
         the_filter = with_filter
@@ -395,15 +441,36 @@ class DOWMLLib:
             # GH-1: This happens when the job failed
             pass
 
+    def _delete_data_assets(self, job_id):
+        job_details = self.get_job_details(job_id, with_contents='names')
+        odr = job_details['entity']['decision_optimization'].get('output_data_references', [])
+        for output in odr:
+            if output.get('type') != 'data_asset':
+                continue
+            if 'location' not in output:
+                self._logger.error(f'Missing \'location\' in details for job {job_id}')
+            elif 'id' not in output['location']:
+                self._logger.error(f'Missing \'location.id\' in details for job {job_id}')
+            else:
+                data_asset_id = output['location']['id']
+                self._logger.debug(f'Deleting data asset {data_asset_id}...')
+                try:
+                    self._client.data_assets.delete(data_asset_id)
+                    self._logger.debug('Done.')
+                except WMLClientError:
+                    self._logger.error('Exception raised while trying to delete the asset', exc_info=True)
+
     def delete_job(self, job_id, hard=False):
         """ Delete the given job
         :param job_id: the job to be deleted
         :param hard: if False, cancel the job. If true, delete it completely
         """
         client = self._get_or_make_client()
+        if hard:
+            self._delete_data_assets(job_id)
         self._logger.debug(f'Deleting job {job_id}...')
         client.deployments.delete_job(job_id, hard)
-        self._logger.debug(f'Done.')
+        self._logger.debug('Done.')
 
     def decode_log(self, output):
         """ Decode the log from DO4WML
@@ -530,7 +597,7 @@ class DOWMLLib:
             return deployment_type
         except KeyError:
             # Something changed. But let's not fail just for that
-            self._logger.warning(f'Error while fetching type of a job!')
+            self._logger.warning('Error while fetching type of a job!')
             return '?????'
 
     def _get_version_from_details(self, job):
@@ -541,13 +608,13 @@ class DOWMLLib:
             model = self._get_model_definition_from_id(model_id)
             deployment_type = model['entity']['wml_model']['type']
             match = re.fullmatch(r"do-....*_([0-9.]*)", deployment_type)
-            version = '?????'
+            engine_version = '?????'
             if match:
-                version = match.group(1)
-            return version
+                engine_version = match.group(1)
+            return engine_version
         except KeyError:
             # Something changed. But let's not fail just for that
-            self._logger.warning(f'Error while fetching version of a job!')
+            self._logger.warning('Error while fetching version of a job!')
             return '?????'
 
     @lru_cache
@@ -570,16 +637,16 @@ class DOWMLLib:
             return size
         except KeyError:
             # Something changed. But let's not fail just for that
-            self._logger.warning(f'Error while fetching size of a job!')
+            self._logger.warning('Error while fetching size of a job!')
             return '?'
 
     def get_jobs(self):
         """Return the list of tuples (status, id, ...) for all jobs in the deployment"""
         client = self._get_or_make_client()
-        self._logger.debug(f'Getting job details...')
+        self._logger.debug('Getting job details...')
         job_details = client.deployments.get_job_details()
-        self._logger.debug(f'Done.')
-        self._logger.debug(f'Getting information about deployments and models...')
+        self._logger.debug('Done.')
+        self._logger.debug('Getting information about deployments and models...')
         result = []
         for job in job_details['resources']:
             status = self._get_job_status_from_details(job)
@@ -587,25 +654,27 @@ class DOWMLLib:
             created = self._get_creation_time_from_details(job, self.tz)
             names = self._get_input_names_from_details(job)
             deployment_type = self._get_type_from_details(job)
-            version = self._get_version_from_details(job)
+            engine_version = self._get_version_from_details(job)
             size = self._get_size_from_details(job)
             JobTuple = namedtuple('Job', ['status', 'id', 'created', 'names', 'type', 'version', 'size'])
             j = JobTuple(status=status, id=job_id, created=created, names=names,
-                         type=deployment_type, version=version, size=size)
+                         type=deployment_type, version=engine_version, size=size)
             result.append(j)
         result.sort(key=attrgetter('created'))
-        self._logger.debug(f'Done.')
+        self._logger.debug('Done.')
         return result
 
     def create_job(self, paths, deployment_id):
         """Create a deployment job (aka a run) and return its id"""
         client = self._get_or_make_client()
         cdd = client.deployments.DecisionOptimizationMetaNames
-        # We always use inline output
+        assert(self.outputs == 'inline' or self.outputs == 'assets')
         cdd_outputdata = cdd.OUTPUT_DATA
+        if self.outputs == 'assets':
+            cdd_outputdata = cdd.OUTPUT_DATA_REFERENCES
         # Assume we use inline data (i.e. content in the job request)
         cdd_inputdata = cdd.INPUT_DATA
-        if not self.inline:
+        if self.inputs == 'assets':
             # But if we don't want inline data, we have to submit
             # input references instead
             cdd_inputdata = cdd.INPUT_DATA_REFERENCES
@@ -621,13 +690,23 @@ class DOWMLLib:
                 {'id': '.*\\.*'}
             ]
         }
+        if self.outputs == 'assets':
+            out = solve_payload[cdd_outputdata][0]
+            out['type'] = 'data_asset'
+            # PyCharm assumes that, because we added a string in the dict on
+            # the previous line, we should only add strings in the same dict.
+            # But this is not how WML does...
+            # noinspection PyTypeChecker
+            out['connection'] = {}
+            # noinspection PyTypeChecker
+            out['location'] = {'name': '${job_id}/${attachment_name}'}
         if self.timelimit:
             params = solve_payload[cdd.SOLVE_PARAMETERS]
             params['oaas.timeLimit'] = 1000 * self.timelimit
         self.create_inputs(paths, cdd_inputdata, solve_payload)
-        self._logger.debug(f'Creating the job...')
-        if self.inline:
-            self._logger.debug(f'Data is inline. Let\'s not print the payload...')
+        self._logger.debug('Creating the job...')
+        if self.inputs == 'inline':
+            self._logger.debug('Data is inline. Let\'s not print the payload...')
         else:
             self._logger.debug(repr(solve_payload))
         dt = datetime.now()
@@ -647,7 +726,7 @@ class DOWMLLib:
             if basename in names:
                 raise SimilarNamesInJob(basename)
             names.append(basename)
-            if self.inline:
+            if self.inputs == 'inline':
                 input_data = {
                     'id': basename,
                     'content': self.get_file_as_data(path)
@@ -686,14 +765,10 @@ class DOWMLLib:
 
     def _get_deployment_id(self):
         """Create deployment if doesn't exist already, return its id"""
-
-        if not self._space_id:
-            self._space_id = self._get_space_id()
-
-        self._logger.debug(f'Getting deployments...')
+        self._logger.debug('Getting deployments...')
         client = self._get_or_make_client()
         deployment_details = client.deployments.get_details()
-        self._logger.debug(f'Done.')
+        self._logger.debug('Done.')
         resources = deployment_details['resources']
         deployment_name = f'{self.DEPLOYMENT_NAME}-{self.model_type}-{self.do_version}-{self.tshirt_size}'
         self._logger.debug(f'Got the list. Looking for deployment named \'{deployment_name}\'')
@@ -701,12 +776,12 @@ class DOWMLLib:
         for r in resources:
             if r['entity']['name'] == deployment_name:
                 deployment_id = r['metadata']['id']
-                self._logger.debug(f'Found it.')
+                self._logger.debug('Found it.')
                 break
         if deployment_id is not None:
             return deployment_id
 
-        self._logger.debug(f'This deployment doesn\'t exist yet. Creating it...')
+        self._logger.debug('This deployment doesn\'t exist yet. Creating it...')
 
         deployment_id = self._create_deployment(deployment_name)
         return deployment_id
@@ -715,7 +790,7 @@ class DOWMLLib:
         # We need a model to create a deployment
         model_id = self._get_model_id()
         # Create the deployment
-        self._logger.debug(f'Creating the deployment itself...')
+        self._logger.debug('Creating the deployment itself...')
         client = self._get_or_make_client()
         cdc = client.deployments.ConfigurationMetaNames
         meta_props = {
@@ -725,16 +800,16 @@ class DOWMLLib:
             cdc.HARDWARE_SPEC: {'name': self.tshirt_size, 'num_nodes': 2}
         }
         deployment = client.deployments.create(artifact_uid=model_id, meta_props=meta_props)
-        self._logger.debug(f'Deployment created.')
+        self._logger.debug('Deployment created.')
         deployment_id = client.deployments.get_id(deployment)
         return deployment_id
 
     def _get_model_id(self):
         """Create an empty model if one doesn't exist, return its id"""
-        self._logger.debug(f'Getting models...')
+        self._logger.debug('Getting models...')
         client = self._get_or_make_client()
         details = client.repository.get_details()
-        self._logger.debug(f'Done.')
+        self._logger.debug('Done.')
         resources = details['models']['resources']
         model_name = f'{self.MODEL_NAME}-{self.model_type}-{self.do_version}'
         self._logger.debug(f'Got the list. Looking for model named \'{model_name}\'')
@@ -742,11 +817,11 @@ class DOWMLLib:
         for r in resources:
             if r['metadata']['name'] == model_name:
                 model_id = r['metadata']['id']
-                self._logger.debug(f'Found it.')
+                self._logger.debug('Found it.')
                 self._logger.debug(f'Model id: {model_id}')
                 break
         if model_id is None:
-            self._logger.debug(f'This model doesn\'t exist yet. Creating it...')
+            self._logger.debug('This model doesn\'t exist yet. Creating it...')
             model_id = self._create_model(model_name)
         return model_id
 
@@ -790,32 +865,16 @@ class DOWMLLib:
                                            meta_props=model_metadata)
         finally:
             os.remove(path)
-        self._logger.debug(f'Model created.')
+        self._logger.debug('Model created.')
         model_id = client.repository.get_model_id(model_details)
-        self._logger.debug(f'Model id: {model_id}')
+        self._logger.debug('Model id: {model_id}')
         return model_id
 
-    def _get_space_id(self):
-        if self._space_id:
-            return self._space_id
-
-        SPACE_ID = _CredentialsProvider.SPACE_ID
-        if SPACE_ID in self._wml_credentials:
-            space_id = self._wml_credentials[SPACE_ID]
-            self._logger.debug(f'Using specified space \'{space_id}\'.')
-        else:
-            space_id = self._find_or_create_space()
-
-        self._logger.debug(f'Setting default space...')
-        self._client.set.default_space(space_id)
-        self._logger.debug(f'Done.')
-        return space_id
-
     def _find_or_create_space(self):
-        """Find the Space to use, create it if it doesn't exist"""
+        """Find the Space to use from its name, create it if it doesn't exist"""
         assert self._client
         client = self._client
-        self._logger.debug(f'Fetching existing spaces...')
+        self._logger.debug('Fetching existing spaces...')
         space_details = client.spaces.get_details()
         resources = space_details['resources']
         self._logger.debug(f'Got the list. Looking for space named \'{self.space_name}\'')
@@ -823,10 +882,10 @@ class DOWMLLib:
         for r in resources:
             if r['entity']['name'] == self.space_name:
                 space_id = r['metadata']['id']
-                self._logger.debug(f'Found it.')
+                self._logger.debug('Found it.')
                 break
         if space_id is None:
-            self._logger.debug(f'This space doesn\'t exist yet. Creating it...')
+            self._logger.debug('This space doesn\'t exist yet. Creating it...')
             # Prepare necessary information
 
             wml_credentials = self._wml_credentials
@@ -854,7 +913,7 @@ class DOWMLLib:
             }
             # Create the space
             space = client.spaces.store(meta_props=metadata)
-            self._logger.debug(f'Space created')
+            self._logger.debug('Space created')
             space_id = client.spaces.get_uid(space)
         self._logger.info(f'Space id: {space_id}')
         return space_id
@@ -899,16 +958,16 @@ class DOWMLLib:
             self._logger.debug(f'Yes, with id {data_asset_id}.')
             if not force:
                 return data_asset_id
-            self._logger.debug(f'Creating new asset with local content.')
+            self._logger.debug('Creating new asset with local content.')
             asset_to_delete = data_asset_id
         else:
-            self._logger.debug(f'No, creating the data asset.')
+            self._logger.debug('No, creating the data asset.')
         data_asset_id = self.create_asset(path, basename)
-        self._logger.debug(f'Done.')
+        self._logger.debug('Done.')
         if asset_to_delete:
-            self._logger.debug(f'Deleting the old data asset.')
+            self._logger.debug('Deleting the old data asset.')
             if self.delete_asset(asset_to_delete):
-                self._logger.debug(f'Done.')
+                self._logger.debug('Done.')
             else:
-                self._logger.warning(f'Could not delete pre-existing asset')
+                self._logger.warning('Could not delete pre-existing asset')
         return data_asset_id
